open Ext
open Printf
open Add

(****************************
 * ADD Variable Elimination
 ****************************)


(* Efficient caching for ADD sizes 
 * Used for multiplying smallest ADDs first in varelim *)
let size_hash = Hashtbl.create 100 
let hsize add = 
  if Hashtbl.mem size_hash add.root then
    Hashtbl.find size_hash add.root
  else begin
    let s = size add in
    Hashtbl.add size_hash add.root s ;
    s
  end 
let cmpsize a b = hsize a - hsize b 

exception OutOfMemory of int

(* Cache defined globally to save time. *)
let multcache = Hashtbl.create 100 

let rec rec_eliminate maxsize avo addset = function
  | [] -> 
      if log_exists log_debug then begin
        dlogf "Final add set:\n";
        List.iter (output (log_stream log_debug)) addset
      end;
      addset
  | v :: l ->
      (* For each variable in the elimination ordering, *)
      (* Select all ADDs with the variable *)
      Timer.start "elim";
      let has_v add = add.vars.(v) in
      let (relevant, irrelevant) = List.partition has_v addset in

      (* Multiply smallest ones first -- WORKS!!! *)
      let multiply x y =
        let product = multiplyc multcache x y in
        if maxsize > 0. && float_of_int(node_size product.root) > maxsize then
          (nlogf "VARIABLE %d FAILED!\n\n" v; raise (OutOfMemory v))
        else
          product in
        (* else (printf "size %d < maxsize %d\n" (node_size product.root) maxsize; product) *) 
      let orelevant = List.sort cmpsize relevant in
      let product = List.fold_left multiply  
        (List.hd orelevant) (List.tl orelevant) in
      Hashtbl.clear multcache;

      (* Pick the factor with fewest vars that aren't in current factor *
      let rec multall rel =
        match (List.sort cmpsize rel) with
          a :: [] -> a
        | a :: b :: l -> multall ((multiply a b) :: l)
        | [] -> assert false in
      let product = multall (indicator_adds.(v) :: relevant) in *)

      (* Always multiply smallest pair of factors next -- Similar
       * to previous method.
      let rec multall rel =
        match (List.sort cmpsize rel) with
          a :: [] -> a
        | a :: b :: l -> multall ((multiply a b) :: l)
        | [] -> assert false in
      let product = multall (indicator_adds.(v) :: relevant) in
      vlogf "Eliminating var %d: %fs\n" v (Timer.delta "elim") ;
       *)

      (* ...and sum out the eliminated variable *)
      let summed_out = sum_out v product in 
      vlogf "Summing out: %fs\n" (Timer.delta "elim") ;
      vlogf "Final size: %d\n" (hsize summed_out); 
      let addset' = summed_out :: irrelevant in
      vlogf "%d factors remaining.\n" (List.length addset');
      if log_exists log_debug then
        List.iter (output (log_stream log_debug)) addset';
      rec_eliminate maxsize avo addset' l 

let eliminate maxsize avo addset lvo =
  rec_eliminate maxsize avo addset lvo

let mn_to_adds mn avo lvo = 
  let numvars = Mn.numvars mn in
  let schema = Mn.schema mn in
  let vn = Circuit.make_vnodes schema in

  (* Create ADDs for each variable *)
  let s_hash = Hashtbl.create 100 in
  let l_hash = Hashtbl.create 100 in
  let indicator_adds = Array.init numvars (indicators_to_add avo lvo vn) in
  let schema = Mn.schema mn in
  let factor_adds = 
    Array.map (factor_to_adds s_hash l_hash avo lvo schema) (Mn.factors mn) in
  (List.flatten (Array.to_list factor_adds), 
   Array.to_list indicator_adds, s_hash, l_hash)

let bn_to_adds bn avo lvo =
  mn_to_adds (Bn.to_mn bn) avo lvo


exception OutOfMemory of int

let rec compile_until_small avo lvo prune_v maxsize adds ind_adds =
  vlogf "cus maxsize = %f\n" maxsize;
  try
    (eliminate maxsize avo (adds @ ind_adds) lvo, adds)
  with OutOfMemory v -> 
    (* Recompute sizes/ordering/etc. *)
    List.iter AddPrune.compute_pruned_sizes adds;
    let a = List.hd adds in
    let avo = a.avo and lvo = a.lvo in
    (* Half the worst-case size of the table with v. *)
    let sizes = AddPrune.max_table_sizes avo lvo adds in
    let adds' = AddPrune.prune_table_to_size prune_v adds (sizes.(v) /. 2.) v in
    compile_until_small avo lvo prune_v maxsize adds' ind_adds


let finish_ac add_roots =
  (* Extract AC roots and multiply them together to get a single root *)
  let extract_ac add =
    let (id, x) = add.root in 
    match x with
      Leaf n -> n
    | _ -> 
      if log_exists log_debug then begin
        output (log_stream log_debug) add
      end;
      raise NonLeafException  in
  let ac_roots = List.map extract_ac add_roots in
  match ac_roots with
     [r] -> r
   | _   -> Circuit.create_times ac_roots 


let build_avo lvo =
  let numvars = List.length lvo in
  let avo = Array.create numvars 0 in
  let rec build_order i = function
    | [] -> ()
    | x :: l -> avo.(x) <- i ; build_order (i+1) l in
  build_order 0 lvo ;
  avo


let varelim prune_one thresh mn = 
  dlogf "Entering varelim...\n";
  Timer.start "ve";
  let lvo = Minfill.minfill_order (Minfill.mn_to_clusters mn) in
  let avo = build_avo lvo in
  nlogf "Minfill ordering: %fs\n" (Timer.delta "ve") ;

  let (factor_adds, ind_adds, s_hash, l_hash) = mn_to_adds mn avo lvo in
  nlogf "ADD creation: %fs\n" (Timer.delta "ve") ;

  let factor_adds = List.map (prune_one s_hash l_hash) factor_adds in
  nlogf "Pruning: %fs\n" (Timer.delta "ve") ;
        
  let add_roots = eliminate 0. avo (factor_adds @ ind_adds) lvo in
  nlogf "Variable elimination: %fs\n" (Timer.delta "ve");
  (finish_ac add_roots, factor_adds)


let varelim_mem prune_v maxsize mn = 
  dlogf "Entering varelim_mem...\n";
  Timer.start "ve";
  let lvo = Minfill.minfill_order (Minfill.mn_to_clusters mn) in
  let avo = build_avo lvo in
  nlogf "Minfill ordering: %fs\n" (Timer.delta "ve") ;

  let (factor_adds, ind_adds, s_hash, l_hash) = mn_to_adds mn avo lvo in
  nlogf "ADD creation: %fs\n" (Timer.delta "ve") ;

  let p = prune_v s_hash l_hash in
  let factor_adds = List.map (AddPrune.prune_one_to_size maxsize p 1.0) factor_adds in
  let (add_roots, factor_adds) = 
    compile_until_small avo lvo p maxsize factor_adds ind_adds in
  nlogf "Variable elimination: %fs\n" (Timer.delta "ve");
  (finish_ac add_roots, factor_adds)


let varelim_choose thresh mn = 
  let prune_one s_hash l_hash (add: sym_add) = 
    let llratios_choose = AddPrune.compute_llratios_choose add in 
    let prh = AddPrune.compute_llratios_prune add.root in
    let root = 
      snd (AddPrune.simplify_add l_hash s_hash llratios_choose prh thresh add) in
    create_add add.avo add.lvo add.vars root in
  varelim prune_one thresh mn


let varelim_prune thresh mn =
  let prune_one s_hash l_hash (add: sym_add) = 
    let llratios = AddPrune.compute_llratios_prune add.root in
    let root = AddPrune.prune_add l_hash s_hash llratios thresh add in
    create_add add.avo add.lvo add.vars root in
  varelim prune_one thresh mn

let varelim_prune_m maxsize mn =
  varelim_mem AddPrune.prune_v maxsize mn

let varelim_choose_m maxsize mn =
  varelim_mem AddPrune.simplify_v maxsize mn
