open Printf
open Ext

(* Globals used for command line parameters *)
let modelfile = ref ""
let outfile = ref ""
let evfile = ref ""
let tofeatures = ref false
let todn = ref false

let usage = "Usage: mconvert -m <model> -o <output>"
let args = Arg.align
 ([("-m", Arg.Set_string modelfile, 
     " Input model file (MN, BIF, XMOD, UAI, or AC)");
   ("-o", Arg.Set_string outfile, " Output model file");
   ("-ev", Arg.Set_string evfile, " Evidence file");
   ("-feat", Arg.Set tofeatures, " Convert MN factors to features");
   ("-dn", Arg.Set todn, 
     " Convert factors to conditional distributions, producing a DN")]
   @ common_arguments)


let make_dn mn =
  let featurelists = Array.create (Array.length mn.Mn.schema) [] in
  let add_to_list f i =
    featurelists.(i) <- f :: featurelists.(i) in
  let add_feature f =
    List.iter (add_to_list f) (Mn.Factor.feature_set_vars [f]) in
  let features = Mn.to_features mn in
  List.iter add_feature features;
  let factors = Array.map (fun f -> Mn.Factor.FeatureSet f) featurelists in
  Mn.create mn.Mn.schema factors
  
let type_to_str = function
  | ACFile -> "circuit"
  | BNFile -> "Bayesian network"
  | MNFile -> "Markov network"
  | UnknownFile -> "unknown"
  | DataFile -> "data"

let main () = 
  Arg.parse args ignore usage;
  if !modelfile = "" || !outfile = "" then
    (Arg.usage args usage; exit 0);

  common_log_init();

  let has_ev = !evfile <> "" in
  let ev = if has_ev then Data.input_example (open_in !evfile)
           else [||] in

  let proc_mn mn =
    let mn = if has_ev then Mn.simplify mn ev else mn in
    let mn = 
      if !todn then make_dn mn 
      else if not !tofeatures then mn 
      else
        let factor_to_fset f = 
          Mn.Factor.FeatureSet (Mn.Factor.to_features f) in
        let factors = Array.map factor_to_fset (Mn.factors mn) in
        Mn.create mn.Mn.schema factors in
    Mn.write_auto !outfile mn in

  if !todn then
    if filetype !outfile = MNFile then 
      tofeatures := true
    else
      nlogf "Ignoring -dn option, which requires a .mn output file.";

  match (filetype !modelfile, filetype !outfile) with
  | (BNFile, BNFile) ->
    let bn = Bn.load_auto !modelfile in
    (* TODO: Convert trees to tables if necessary... *)
    let bn = if has_ev then Bn.simplify bn ev else bn in
    Bn.write_auto !outfile bn
  | (BNFile, MNFile) ->
    let bn = Bn.load_auto !modelfile in
    proc_mn (Bn.to_mn bn)
  | (MNFile, MNFile) ->
    proc_mn (Mn.load_auto !modelfile)
  | (ACFile, MNFile) -> 
    let (ac, fl) = Circuit.load_with_features (open_in !modelfile) in
    if has_ev then Circuit.prune_for_evidence ac ev;
    let f_to_factor f =
      Mn.Factor.Feature
        {Mn.Factor.cond = f.Circuit.cond;
         Mn.Factor.weight_id = (Circuit.feature_node f).Circuit.id;
         Mn.Factor.weight = log f.Circuit.weight} in
    let factors = Array.of_list (List.map f_to_factor fl) in
    let mn = Mn.create (ac.Circuit.schema) factors in
    proc_mn mn
  | (ACFile, ACFile) -> 
    let (ac, fl) = Circuit.load_with_features (open_in !modelfile) in
    if has_ev then Circuit.prune_for_evidence ac ev;
    let fl = if has_ev then Circuit.prune_features fl ev else fl in
    Circuit.output_with_features (open_out !outfile) ac fl
  | (t1, t2) -> nlogf "ERROR: Converting from %s to %s not supported.\n"
          (type_to_str t1) (type_to_str t2)

;;
main ()
